import numpy as np
from time import sleep
import configparser

from pymeasure.experiment import Procedure, BooleanParameter, FloatParameter, IntegerParameter

from ..classes.stage import Stage
from ..classes.dac import DAC
from ..classes.hall_sensor import HallSensor
from ..classes.logger import log

class B_Sweep(Procedure):
    config_file = "src/configs/procedures_inputs.ini"
    name = "B-Sweep"

    config = configparser.ConfigParser()
    config.read(config_file)

    if name in config:
        section = config[name]
    else:
        section = dict()
    
    acq_time = FloatParameter('Aquisition time', units='s', default=section.get("acq_time", 1))
    freq = FloatParameter('Aquisition Frequency', units='Hz', default=section.get("freq", 50))
    b_min = FloatParameter('From B (max 6A)', units='A', minimum=-6., maximum=6., default=section.get("b_min", 0))
    b_max = FloatParameter('To B (max 6A)', units='A', minimum=-6., maximum=6., default=section.get("b_max", 5))
    b_step = FloatParameter('Step', units='A', default=section.get("b_step", 0.1))
    x = FloatParameter('Position x', units='mm', default=section.get("x", 0.))
    y = FloatParameter('Position y', units='mm', default=section.get("y", 0))
    loop = BooleanParameter('Loop back for hysteresis', default=section.get("loop", True))

    DATA_COLUMNS = ['Iteration', 'X Position', 'Y Position', 'Voltage', 'Magnetic Field [A]', 'Magnetic Field [mT]']

    def startup(self):
        log.info("Setting up the data aquisition")
        self.save_config()

    def save_config(self):
        config = configparser.ConfigParser()
        config[self.name] = {
            "acq_time"  : str(self.acq_time), 
            "freq"      : str(self.freq), 
            "b_min"     : str(self.b_min), 
            "b_max"     : str(self.b_max), 
            "b_step"    : str(self.b_step), 
            "x"         : str(self.x), 
            "y"         : str(self.y), 
            "loop"      : str(self.loop), 
        }
        with open(self.config_file, "w") as config_file:
            config.write(config_file)

    def execute(self):

        voltage = DAC("ai0")
        field   = DAC("ao1")     
        hall_sensor = HallSensor()

        b_values = np.linspace(self.b_min, self.b_max, int(np.abs(self.b_max-self.b_min)//self.b_step + 1), endpoint=True)
        
        if (self.loop):
            b_values = np.concatenate((b_values, np.flip(b_values)[1:]))

        if (self.b_min != 0):
            log.info(f"Setting up magnetic field to {self.b_min}A, wait 1s")
            field.write(self.b_min)
            sleep(1)

        try:
            stage = Stage()
            stage.enable_connections()
        except:
            log.info("Close connection with motors before running an experiment, motors can not be used")

        log.info(f"Move stage to ({self.x}mm, {self.y}mm)")
        stage.move_x_to(self.x)
        stage.move_y_to(self.y)
        stage.wait_stable()

        if (int(round(self.acq_time * self.freq)) <= 2):
            log.info(f"Only {int(round(self.acq_time * self.freq))} points used for the average! Start voltage aquisition.")
        else:
            log.info("Start voltage aquisition")
        
        # Start sweeping loop
        for i in range(len(b_values)):
            
            field.write(b_values[i])

            average_voltage = np.arange(0, int(round(self.acq_time * self.freq)), dtype='float64')

            for step in range(len(average_voltage)):
                average_voltage[step] = voltage.read()
                sleep(1/self.freq)

            data = {
                'Iteration': i,
                'X Position': stage.get_x_pos(),
                'Y Position': stage.get_y_pos(),
                'Magnetic Field [A]': b_values[i],
                'Magnetic Field [mT]': hall_sensor.read_mT(),
                'Voltage': np.mean(average_voltage)
            }

            log.debug("Produced numbers: %s" % data)
            self.emit('results', data)
            self.emit('progress', 100 * i / len(b_values))
            if self.should_stop():
                break

        field.write(0.)
        stage.close_connections()

    def shutdown(self):
        log.info("Turning off connection with motors and magnetic field")
        field   = DAC("ao1")     
        field.write(0)

        sleep(0.4)

        try:
            stage = Stage()
            stage.close_connections()
        except:
            pass
        log.info("Finished acquisition")
